from collections import Counter
import datetime
from typing import Any, List

from typing import Tuple

from .config import KNOWN_IMPHASHES

#import yarobot_rs
import os
import re
import logging


from yarobot.yarobot_rs import ScoringEngine, TokenInfo, TokenType


def get_uint_string(magic):
    print(magic)
    return f"uint16(0) == 0x{hex(magic[1])[2:]}{hex(magic[0])[2:]}"


def sanitize_rule_name(path: str, file: str) -> str:
    """Generate a valid YARA rule name from path and filename.

    - Prefix with folder name if too short
    - Ensure it doesn't start with a number
    - Replace invalid chars with underscore
    - De-duplicate underscores
    """
    file_base = os.path.splitext(file)[0]
    cleaned = file_base
    if len(file_base) < 8:
        cleaned = path.split("\\")[-1:][0] + "_" + cleaned
    if re.search(r"^[0-9]", cleaned):
        cleaned = "sig_" + cleaned
    cleaned = re.sub(r"[^\w]", "_", cleaned)
    cleaned = re.sub(r"_+", "_", cleaned)
    return cleaned


def get_timestamp_basic(date_obj=None):
    return (
        date_obj.strftime("%Y-%m-%d")
        if date_obj
        else datetime.datetime.now().strftime("%Y-%m-%d")
    )


def get_file_range(size, fm_size):
    size_string = ""
    # max sample size - args.filesize_multiplier times the original size
    max_size_b = size * fm_size
    # Minimum size
    if max_size_b < 1024:
        max_size_b = 1024
    # in KB
    max_size = int(max_size_b / 1024)
    max_size_kb = max_size
    # Round
    if len(str(max_size)) == 2:
        max_size = int(round(max_size, -1))
    elif len(str(max_size)) == 3:
        max_size = int(round(max_size, -2))
    elif len(str(max_size)) == 4:
        max_size = int(round(max_size, -3))
    elif len(str(max_size)) >= 5:
        max_size = int(round(max_size, -3))
    size_string = f"filesize < {max_size}KB"
    logging.getLogger("yarobot").debug(
        "File Size Eval: SampleSize (b): %s SizeWithMultiplier (b/Kb): %s / %s RoundedSize: %s",
        str(size),
        str(max_size_b),
        str(max_size_kb),
        str(max_size),
    )
    return size_string


def generate_general_condition(file_info, nofilesize, filesize_multiplier, args):
    """
    Generates a general condition for a set of files
    :param file_info:
    :return:
    """
    conditions = []
    pe_module_neccessary = False

    # Different Magic Headers and File Sizes
    magic_headers = []
    file_sizes = []
    imphashes = []

    for filePath in file_info:
        if not file_info[filePath].magic:
            continue
        magic = file_info[filePath].magic
        size = file_info[filePath].size
        imphash = file_info[filePath].imphash

        # Add them to the lists
        if magic not in magic_headers and magic != "":
            magic_headers.append(magic)
        if size not in file_sizes:
            file_sizes.append(size)
        if imphash not in imphashes and imphash != "":
            imphashes.append(imphash)

    # If different magic headers are less than 5
    if len(magic_headers) <= 5:
        magic_string = " or ".join(get_uint_string(h) for h in magic_headers)
        if " or " in magic_string:
            conditions.append("( {0} )".format(magic_string))
        else:
            conditions.append("{0}".format(magic_string))

    # Biggest size multiplied with maxsize_multiplier
    if not nofilesize and len(file_sizes) > 0:
        conditions.append(get_file_range(max(file_sizes), filesize_multiplier))

    # If different magic headers are less than 5
    if len(imphashes) == 1 and not args.noextras:
        conditions.append('pe.imphash() == "{0}"'.format(imphashes[0]))
        pe_module_neccessary = True

    # If enough attributes were special
    condition_string = " and ".join(conditions)

    return condition_string, pe_module_neccessary


class RuleGenerator:
    def __init__(self, args):
        self.prefix = args.prefix
        self.author = args.author
        self.ref = args.ref

    def generate_rule(
        self, rule_name, file, hashes, rule_strings, rule_opcodes, conditions
    ):
        # Print rule title
        rule = (
            f"rule {rule_name} {{\n"
            f"\tmeta:\n"
            f'\t\tdescription = "{self.prefix} - file {file}"\n'
            f'\t\tauthor = "{self.author}"\n'
            f'\t\treference = "{self.ref}"\n'
            f'\t\tdate = "{get_timestamp_basic()}"\n'
        )
        for i, hash in enumerate(hashes):
            rule += f'\t\thash{i + 1} = "{hash}"\n'
        rule += "\tstrings:\n"
        rule += "\n".join(rule_strings)
        rule += "\n"
        rule += "\n".join(rule_opcodes)
        rule += "\n\tcondition:\n"
        rule += "\t\t%s\n" % conditions
        rule += "}\n\n"
        return rule


def add_conditions(
    conditions,
    subconditions,
    rule_strings,
    rule_opcodes,
    high_scoring_strings,
    pe_conditions_add,
):
    # String combinations
    cond_op = ""  # opcodes condition
    cond_hs = ""  # high scoring strings condition
    cond_ls = ""  # low scoring strings condition

    low_scoring_strings = len(rule_strings) - high_scoring_strings
    if high_scoring_strings > 0:
        cond_hs = "1 of ($x*)"
    if low_scoring_strings > 0:
        if low_scoring_strings > 10:
            if high_scoring_strings > 0:
                cond_ls = "4 of them"
            else:
                cond_ls = "8 of them"
        else:
            cond_ls = "all of them"

    # If low scoring and high scoring
    cond_combined = "all of them"
    needs_brackets = False
    if low_scoring_strings > 0 and high_scoring_strings > 0:
        # If PE conditions have been added, don't be so strict with the strings
        if pe_conditions_add:
            cond_combined = "{0} or {1}".format(cond_hs, cond_ls)
            needs_brackets = True
        else:
            cond_combined = "{0} and {1}".format(cond_hs, cond_ls)
    elif low_scoring_strings > 0 and not high_scoring_strings > 0:
        cond_combined = "{0}".format(cond_ls)
    elif not low_scoring_strings > 0 and high_scoring_strings > 0:
        cond_combined = "{0}".format(cond_hs)
    if rule_opcodes:
        cond_op = " and all of ($op*)"
        # Opcodes (if needed)
    if cond_op or needs_brackets:
        subconditions.append("( {0}{1} )".format(cond_combined, cond_op))
    else:
        subconditions.append(cond_combined)


def generate_rule_tokens(
    scoring_engine, args, strings, utf16strings, opcodes, opcode_num
) -> Tuple[List[str], int, List[str]]:
    # Rule String generation
    (
        rule_strings,
        high_scoring_strings,
    ) = scoring_engine.generate_rule_strings(
        args.score,
        args.high_scoring,
        args.strings_per_rule,
        (strings or []) + (utf16strings or []),
    )  # generate_rule_strings(args,(strings or []) + (utf16strings or []),)

    rule_opcodes = []
    if opcodes:
        rule_opcodes = generate_rule_opcodes(opcodes, opcode_num)
    return rule_strings, high_scoring_strings, rule_opcodes


def generate_simple_rule(
    rg: RuleGenerator,
    printed_rules,
    args,
    scoring_engine: ScoringEngine,
    strings: List[TokenInfo] | None,
    opcodes: List[TokenInfo] | None,
    utf16strings: List[TokenInfo] | None,
    info,
    fname,
) -> str:
    if not strings and not utf16strings:
        logging.getLogger("yarobot").warning(
            "[W] Not enough high scoring strings to create a rule. (Try -z 0 to reduce the min score or --opcodes to include opcodes) FILE: %s",
            fname,
        )
        return False
    # Skip if there is nothing to do
    logging.getLogger("yarobot").info(
        "[+] Generating rule for %s, %d strings, %d opcodes, %d utf16strs",
        fname,
        len(strings),
        len(opcodes),
        len(utf16strings),
    )

    # Print rule title ----------------------------------------

    (path, file) = os.path.split(fname)
    # Prepare name via helper
    cleanedName = sanitize_rule_name(path, file)
    # Check if already printed
    if cleanedName in printed_rules:
        printed_rules[cleanedName] += 1
        cleanedName = cleanedName + "_" + str(printed_rules[cleanedName])
    else:
        printed_rules[cleanedName] = 1

    # Condition -----------------------------------------------
    # Conditions list (will later be joined with 'or')
    conditions = []  # AND connected
    subconditions = []  # OR connected

    # Condition PE
    # Imphash and Exports - applicable to PE files only
    condition_pe = []
    condition_pe_part1 = []
    condition_pe_part2 = []

    def add_extras():
        # Add imphash - if certain conditions are met
        if info.imphash not in scoring_engine.good_imphashes_db and info.imphash != "":
            # Comment to imphash
            imphash = info.imphash
            comment = ""
            if imphash in KNOWN_IMPHASHES:
                comment = " /* {0} */".format(KNOWN_IMPHASHES[imphash])
            # Add imphash to condition
            condition_pe_part1.append(
                'pe.imphash() == "{0}"{1}'.format(imphash, comment)
            )
            pe_module_necessary = True
        if info.exports:
            e_count = 0
            for export in info.exports:
                if export not in scoring_engine.good_exports_db:
                    condition_pe_part2.append('pe.exports("{0}")'.format(export))
                    e_count += 1
                    pe_module_necessary = True
                if e_count > 5:
                    break

    if not args.noextras and info.magic.startswith(b"MZ"):
        add_extras()

    # 1st Part of Condition 1
    def add_basic_conditions(conditions, info, args):
        basic_conditions: List[Any] = []
        # Filesize
        if not args.nofilesize:
            basic_conditions.insert(
                0, get_file_range(info.size, args.filesize_multiplier)
            )
        # Magic
        if info.magic != b"":
            uint_string = get_uint_string(info.magic)
            basic_conditions.insert(0, uint_string)
        conditions.append(" and ".join(basic_conditions))

    add_basic_conditions(conditions, info, args)
    # Add extra PE conditions to condition 1
    pe_conditions_add = False
    if condition_pe_part1 or condition_pe_part2:
        if len(condition_pe_part1) == 1:
            condition_pe.append(condition_pe_part1[0])
        elif len(condition_pe_part1) > 1:
            condition_pe.append(f"( {' or '.join(condition_pe_part1)} )")
        if len(condition_pe_part2) == 1:
            condition_pe.append(condition_pe_part2[0])
        elif len(condition_pe_part2) > 1:
            condition_pe.append(f"({' and '.join(condition_pe_part2)} )")
        # Marker that PE conditions have been added
        pe_conditions_add = True
        # Add to sub condition
        subconditions.append(" and ".join(condition_pe))

    rule_strings, high_scoring_strings, rule_opcodes = generate_rule_tokens(
        scoring_engine, args, strings, utf16strings, opcodes, args.opcode_num
    )
    add_conditions(
        conditions,
        subconditions,
        rule_strings,
        rule_opcodes,
        high_scoring_strings,
        pe_conditions_add,
    )

    # Now add string condition to the conditions
    if len(subconditions) == 1:
        conditions.append(subconditions[0])
    elif len(subconditions) > 1:
        conditions.append("( %s )" % " or ".join(subconditions))

    # Create condition string
    condition_string = " and\n      ".join(conditions)

    return rg.generate_rule(
        cleanedName, file, [info.sha256], rule_strings, rule_opcodes, condition_string
    )


def generate_super_rule(
    rg: RuleGenerator,
    super_rule,
    infos,
    args,
    scoring_engine,
    printed_rules,
    super_rule_names,
    printed_combi,
    super_rule_count,
    opcodes,
):
    # Prepare Name
    rule_name = ""
    file_list = []
    hashes = []
    # Loop through files
    print("Generating super rule for %s" % super_rule)
    imphashes = Counter()
    for filePath in super_rule.files:
        (path, file) = os.path.split(filePath)
        file_list.append(file)
        # Prepare name via helper
        cleanedName = sanitize_rule_name(path, file)
        # Append it to the full name
        rule_name += "_" + cleanedName
        # Check if imphash of all files is equal
        imphash = infos[filePath].imphash
        hashes.append(infos[filePath].sha256)
        if imphash != "-" and imphash != "":
            imphashes.update([imphash])

    # Imphash usable
    if len(imphashes) == 1:
        unique_imphash = list(imphashes.items())[0][0]
        if unique_imphash in scoring_engine.good_imphashes_db:
            unique_imphash = ""

    # Shorten rule name
    rule_name = rule_name[:124]
    # Add count if rule name already taken
    if rule_name not in super_rule_names:
        rule_name = "%s_%s" % (rule_name, super_rule_count)
    super_rule_names.append(rule_name)

    # File name starts with a number
    if re.search(r"^[0-9]", rule_name):
        rule_name = "sig_" + rule_name
    # clean name from all characters that would cause errors
    rule_name = re.sub(r"[^\w]", "_", rule_name)
    # Check if already printed
    if rule_name in printed_rules:
        printed_combi[rule_name] += 1
        rule_name = rule_name + "_" + str(printed_combi[rule_name])
    else:
        printed_combi[rule_name] = 1

    rule_strings, high_scoring_strings, rule_opcodes = generate_rule_tokens(
        scoring_engine, args, super_rule.strings, [], opcodes, args.opcode_num
    )

    # Condition -----------------------------------------------
    # Conditions list (will later be joined with 'or')
    conditions = []
    subconditions = []
    # 1st condition
    # Evaluate the general characteristics
    file_info_super = {}
    pe_module_necessary = False
    for filePath in super_rule.files:
        file_info_super[filePath] = infos[filePath]
    condition_strings, pe_module_necessary_gen = generate_general_condition(
        infos, args.nofilesize, args.filesize_multiplier, args
    )
    if pe_module_necessary_gen:
        pe_module_necessary = True

    # 2nd condition
    # String combinations
    add_conditions(
        conditions,
        subconditions,
        rule_strings,
        rule_opcodes,
        high_scoring_strings,
        pe_module_necessary,
    )
    # Now add string condition to the conditions
    if len(subconditions) == 1:
        conditions.append(subconditions[0])
    elif len(subconditions) > 1:
        conditions.append("( %s )" % " or ".join(subconditions))
    # Create condition string
    condition_string = "\n      ) or ( ".join(conditions)

    return rg.generate_rule(
        rule_name,
        ", ".join(file_list),
        hashes,
        rule_strings,
        rule_opcodes,
        condition_string,
    )


def generate_top_info(author, identifier, ref, license):
    # General Info
    general_info = "/*\n"
    general_info += "   YARA Rule Set\n"
    general_info += f"   Author: {author}\n"
    general_info += f"   Date: {get_timestamp_basic()}\n"
    general_info += f"   Identifier: {identifier}\n"
    general_info += f"   Reference: {ref}\n"
    if license != "":
        general_info += f"   License: {license}\n"
    general_info += "*/\n\n"
    return general_info


def generate_rules(
    rg: RuleGenerator,
    scoring_engine,
    args,
    file_strings,
    file_opcodes,
    file_utf16strings,
    super_rules,
    opcode_super_rules,
    utf16_super_rules,
    file_info,
):
    fdata = ""
    with open(args.output_rule_file, "w") as fh:
        fdata = generate_top_info(args.author, args.identifier, args.ref, args.license)

        # GLOBAL RULES ----------------------------------------------------
        if args.globalrule:
            condition, pe_module_necessary = generate_general_condition(
                file_info, args.nofilesize, args.filesize_multiplier, args
            )

            # Global Rule
            if condition != "":
                global_rule = (
                    "/* Global Rule -------------------------------------------------------------- */\n"
                    "/* Will be evaluated first, speeds up scanning process, remove at will */\n\n"
                    "global private rule gen_characteristics {\n"
                    "\tcondition:\n"
                    f"\t\t{condition}\n}}\n\n"
                )

                fdata += global_rule

        # General vars
        rules = ""
        printed_rules = {}
        rule_count = 0
        super_rule_count = 0

        # PROCESS SIMPLE RULES ----------------------------------------------------
        logging.getLogger("yarobot").info("[+] Generating Simple Rules ...")

        # logging.getLogger("yarobot").info(file_strings)

        # GENERATE SIMPLE RULES -------------------------------------------
        fdata += "/* Rule Set ----------------------------------------------------------------- */\n\n"
        all_files_set = set(file_strings.keys())
        all_files_set.update(file_opcodes.keys())
        all_files_set.update(file_utf16strings.keys())

        for filePath in all_files_set:
            if rule := generate_simple_rule(
                rg,
                printed_rules,
                args,
                scoring_engine,
                file_strings[filePath] if filePath in file_strings.keys() else [],
                file_opcodes[filePath] if filePath in file_opcodes.keys() else [],
                file_utf16strings[filePath]
                if filePath in file_utf16strings.keys()
                else [],
                file_info[filePath],
                filePath,
            ):
                rules += rule
                rule_count += 1

        # GENERATE SUPER RULES --------------------------------------------
        if not args.nosuper:
            rules += "/* Super Rules ------------------------------------------------------------- */\n\n"
            super_rule_names = []

            print("[+] Generating Super Rules ...")
            printed_combi = {}
            for super_rule in super_rules + opcode_super_rules + utf16_super_rules:
                rules += generate_super_rule(
                    rg,
                    super_rule,
                    file_info,
                    args,
                    scoring_engine,
                    printed_rules,
                    super_rule_names,
                    printed_combi,
                    super_rule_count,
                    None,
                )
                super_rule_count += 1

        # WRITING RULES TO FILE
        # PE Module -------------------------------------------------------
        if not args.noextras:
            if "pe." in rules:
                fdata += 'import "pe"\n\n'
        # RULES ------------------------------
        fdata += rules
        fh.write(fdata)
        # Print rules to command line -------------------------------------
        logging.getLogger("yarobot").debug(rules)

    return (rule_count, super_rule_count, fdata)


def generate_string_repr(is_super_string, i, stringe):
    return f'\t\t${"x" if is_super_string else "s"}{i + 1} = "{stringe.reprz.replace("\\", "\\\\").replace('"', '\\"')}" {"wide" if stringe.typ == TokenType.UTF16LE else "ascii"}\
 {"fullword" if stringe.fullword else ""} /*{stringe.notes}*/'


def generate_opcode_repr(i, opcode):
    return f"\t\t$op{i} = {{{opcode.reprz}}}"


def generate_rule_opcodes(opcode_elements, opcodes_per_rule):
    # Adding the opcodes --------------------------------------
    rule_opcodes = []
    for i, opcode in enumerate(opcode_elements):
        rule_opcodes.append(generate_opcode_repr(i, opcode))
        if i >= opcodes_per_rule:
            break
    return rule_opcodes


def generate_rule_strings(
    scoring_engine, args, string_elements
) -> Tuple[List[str], int]:
    rule_strings = []
    # Adding the strings --------------------------------------

    # string_elements = list(set(string_elements))
    string_elements = sorted(string_elements, key=lambda x: x.score, reverse=True)
    high_scoring_strings = 0
    for i, stringe in enumerate(string_elements):
        # Collect the data
        string = stringe.reprz

        if string in scoring_engine.good_strings_db:
            stringe.add_note(
                f"goodware string - occured {scoring_engine.good_strings_db[string]} times"
            )

        if args.score:
            stringe.add_note(f" / score: {stringe.score} /")
        else:
            logging.getLogger("yarobot").debug("NO SCORE: %s", string)

        if stringe.b64:
            stringe.add_note(
                f" / base64 encoded string '{scoring_engine.base64strings[string]}' /"
            )
        if stringe.hexed:
            stringe.add_note(
                f" / hex encoded string '{yarobot_rs.remove_non_ascii_drop(scoring_engine.hex_enc_strings[string]).decode()}' /"
            )
        if stringe.from_pestudio and args.score:
            stringe.add_note(f" / PEStudio Blacklist: {args.pestudio_marker[string]} /")
        if stringe.reversed:
            stringe.add_note(
                f" / reversed goodware string '{scoring_engine.reversed_strings[string]}' /"
            )

        is_super_string = float(stringe.score) > args.high_scoring
        if is_super_string:
            high_scoring_strings += 1
        rule_strings.append(generate_string_repr(is_super_string, i, stringe))

        # If too many string definitions found - cut it at the
        # count defined via command line param -rc
        if (i + 1) >= args.strings_per_rule:
            break

    else:
        logging.getLogger("yarobot").info(
            "[-] Not enough unique opcodes found to include them"
        )

    return rule_strings, high_scoring_strings



import gzip
import os

import orjson as json

from .config import DB_PATH


def load(filename, just_json=False):
    if just_json:
        with open(filename, "rb") as file:
            return json.loads(file.read())
    else:
        file = gzip.GzipFile(filename, "rb")
        object = json.loads(file.read())
        file.close()
        return object


def get_abs_path(filename):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), filename)


def save(object, filename):
    path = os.path.join(DB_PATH, filename)
    with open(path, "wb") as file:
        file.write(bytes(json.dumps(object)))


def load_db(file, local_counter, just_json=False):
    filePath = os.path.join(DB_PATH, file)
    print("[+] Loading %s ..." % filePath)
    before = len(local_counter)
    js = load(get_abs_path(filePath), just_json)
    local_counter.update(js)
    added = len(local_counter) - before
    print("[+] Total: %s / Added %d entries" % (len(local_counter), added))

    return len(local_counter)



#!/usr/bin/env python


"""
yarGen - Yara Rule Generator, Copyright (c) 2015, Florian Roth
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the copyright owner nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL Florian Roth BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import os
import sys
import logging

import time
from collections import Counter
from lxml import etree

from .common import get_abs_path, load, load_db
from .config import DB_PATH, PE_STRINGS_FILE

import pstats

import cProfile
from .rule_generator import RuleGenerator, generate_rules

# from app.scoring import extract_stats_by_file, sample_string_evaluation
from .config import RELEVANT_EXTENSIONS

from  yarobot import yarobot_rs

import click
import os
import sys

from . import database


def getPrefix(prefix, identifier):
    """
    Get a prefix string for the rule description based on the identifier
    :param prefix:
    :param identifier:
    :return:
    """
    if prefix == "Auto-generated rule":
        return identifier
    else:
        return prefix


def getIdentifier(id, path):
    """
    Get a identifier string - if the provided string is the path to a text file, then read the contents and return it as
    reference, otherwise use the last element of the full path
    :param ref:
    :return:
    """
    # Identifier
    if id == "not set" or not os.path.exists(id):
        # Identifier is the highest folder name
        return os.path.basename(path.rstrip("/"))
    else:
        # Read identifier from file
        identifier = open(id).read()
        print("[+] Read identifier from file %s > %s" % (id, identifier))
        return identifier


def getReference(ref):
    """
    Get a reference string - if the provided string is the path to a text file, then read the contents and return it as
    reference
    :param ref:
    :return:
    """
    if os.path.exists(ref):
        reference = open(ref).read()
        print("[+] Read reference from file %s > %s" % (ref, reference))
        return reference
    else:
        return ref


def emptyFolder(dir):
    """
    Removes all files from a given folder
    :return:
    """
    for file in os.listdir(dir):
        filePath = os.path.join(dir, file)
        try:
            if os.path.isfile(filePath):
                print("[!] Removing %s ..." % filePath)
                os.unlink(filePath)
        except Exception as e:
            print(e)


def initialize_pestudio_strings():
    # if not os.path.isfile(get_abs_path(PE_STRINGS_FILE)):
    #    return None
    print("[+] Processing PEStudio strings ...")

    pestudio_strings = {}

    tree = etree.parse(get_abs_path(PE_STRINGS_FILE))
    processed_strings = {}
    pestudio_strings["strings"] = tree.findall(".//string")
    pestudio_strings["av"] = tree.findall(".//av")
    pestudio_strings["folder"] = tree.findall(".//folder")
    pestudio_strings["os"] = tree.findall(".//os")
    pestudio_strings["reg"] = tree.findall(".//reg")
    pestudio_strings["guid"] = tree.findall(".//guid")
    pestudio_strings["ssdl"] = tree.findall(".//ssdl")
    pestudio_strings["ext"] = tree.findall(".//ext")
    pestudio_strings["agent"] = tree.findall(".//agent")
    pestudio_strings["oid"] = tree.findall(".//oid")
    pestudio_strings["priv"] = tree.findall(".//priv")
    for category, elements in pestudio_strings.items():
        for elem in elements:
            processed_strings[elem.text] = (5, category)
    return processed_strings


def load_databases():
    good_strings_db = Counter()
    good_opcodes_db = Counter()
    good_imphashes_db = Counter()
    good_exports_db = Counter()

    # Initialize all databases
    for file in os.listdir(get_abs_path(DB_PATH)):
        if file.endswith(".db") or file.endswith(".json"):
            if file.startswith("good-strings"):
                load_db(
                    file, good_strings_db, True if file.endswith(".json") else False
                )
            if file.startswith("good-opcodes"):
                load_db(
                    file, good_opcodes_db, True if file.endswith(".json") else False
                )
            if file.startswith("good-imphashes"):
                pass  # load_db(file, good_imphashes_db, True if file.endswith(".json") else False) TODO
            if file.startswith("good-exports"):
                pass  # load_db(file, good_exports_db, True if file.endswith(".json") else False) TODO
    return good_strings_db, good_opcodes_db, good_imphashes_db, good_exports_db


def process_folder(
    args,
    folder,
    good_strings_db={},
    good_opcodes_db={},
    good_imphashes_db={},
    good_exports_db={},
    pestudio_strings={},
):
    if args.opcodes and len(good_opcodes_db) < 1:
        logging.getLogger("yarobot").warning(
            "Missing goodware opcode databases.    Please run 'yarobot update' to retrieve the newest database set."
        )
        args.opcodes = False

    if len(good_exports_db) < 1 and len(good_imphashes_db) < 1:
        logging.getLogger("yarobot").warning(
            "Missing goodware imphash/export databases.     Please run 'yarobot update' to retrieve the newest database set."
        )

    if len(good_strings_db) < 1:
        logging.getLogger("yarobot").warning(
            "no goodware databases found.     Please run 'yarobot update' to retrieve the newest database set."
        )
        # sys.exit(1)

    # Scan malware files
    logging.getLogger("yarobot").info(f"[+] Generating YARA rules from {folder}")
    (
        combinations,
        super_rules,
        utf16_combinations,
        utf16_super_rules,
        opcode_combinations,
        opcode_super_rules,
        file_strings,
        file_opcodes,
        file_utf16strings,
        file_info,
        scoring_engine,
    ) = yarobot_rs.process_malware(
        folder,
        args.recursive,
        RELEVANT_EXTENSIONS,
        args.min_size,
        args.max_size,
        args.max_file_size,
        args.opcodes,
        args.debug,
        args.excludegood,
        args.min_score,
        args.superrule_overlap,
        good_strings_db,
        good_opcodes_db,
        good_imphashes_db,
        good_exports_db,
        pestudio_strings,
    )
    # Apply intelligent filters
    logging.getLogger("yarobot").info(
        "[-] Applying intelligent filters to string findings ..."
    )
    file_strings = {
        fpath: scoring_engine.filter_string_set(strings)
        for fpath, strings in file_strings.items()
    }

    file_opcodes = {
        fpath: scoring_engine.filter_opcode_set(opcodes)
        for fpath, opcodes in file_opcodes.items()
    }

    file_utf16strings = {
        fpath: scoring_engine.filter_string_set(utf16strings)
        for fpath, utf16strings in file_utf16strings.items()
    }

    # Create Rule Files
    rg = RuleGenerator(args)
    (rule_count, super_rule_count, rules) = generate_rules(
        rg,
        scoring_engine,
        args,
        file_strings,
        file_opcodes,
        file_utf16strings,
        super_rules,
        opcode_super_rules,
        utf16_super_rules,
        file_info,
    )

    print("[=] Generated %s SIMPLE rules." % str(rule_count))
    if not args.nosuper:
        print("[=] Generated %s SUPER rules." % str(super_rule_count))
    print("[=] All rules written to %s" % args.output_rule_file)
    return rules

@click.group()
def cli():
    pass

@cli.command()
@click.argument("malware_path", type=click.Path(exists=True))
@click.option(
    "-y",
    "--min-size",
    help="Minimum string length to consider (default=8)",
    type=int,
    default=8,
)
@click.option(
    "-z",
    "--min-score",
    help="Minimum score to consider (default=5)",
    type=int,
    default=5,
)
@click.option(
    "-x",
    "--high-scoring",
    help='Score required to set string as "highly specific string" (default: 30)',
    type=int,
    default=30,
)
@click.option(
    "-w",
    "--superrule-overlap",
    help="Minimum number of strings that overlap to create a super rule (default: 5)",
    type=int,
    default=5,
)
@click.option(
    "-s",
    "--max-size",
    help="Maximum length to consider (default=128)",
    type=int,
    default=128,
)
@click.option(
    "-rc",
    "--strings-per-rule",
    help="Maximum number of strings per rule (default=15, intelligent filtering will be applied)",
    type=int,
    default=15,
)
@click.option(
    "--excludegood",
    help="Force the exclude all goodware strings",
    is_flag=True,
    default=False,
)
@click.option(
    "-o", "--output-rule-file", help="Output rule file", default="yarobot_rules.yar"
)
@click.option(
    "-e", "--output-dir-strings", help="Output directory for string exports", default=""
)
@click.option("-a", "--author", help="Author Name", default="yarobot Rule Generator")
@click.option(
    "--ref",
    help="Reference (can be string or text file)",
    default="https://github.com/ogre2007/yarobot",
)
@click.option("-l", "--license", help="License", default="")
@click.option(
    "-p",
    "--prefix",
    help="Prefix for the rule description",
    default="Auto-generated rule",
)
@click.option(
    "-b",
    "--identifier",
    help="Text file from which the identifier is read (default: last folder name in the full path)",
    default="not set",
)
@click.option(
    "--score",
    help="Show the string scores as comments in the rules",
    is_flag=True,
    default=False,
)
@click.option(
    "--nosimple",
    help="Skip simple rule creation for files included in super rules",
    is_flag=True,
    default=False,
)
@click.option(
    "--nomagic",
    help="Don't include the magic header condition statement",
    is_flag=True,
    default=False,
)
@click.option(
    "--nofilesize",
    help="Don't include the filesize condition statement",
    is_flag=True,
    default=False,
)
@click.option(
    "-fm",
    "--filesize-multiplier",
    help="Multiplier for the maximum 'filesize' condition value (default: 2)",
    type=int,
    default=2,
)
@click.option(
    "--globalrule",
    help="Create global rules (improved rule set speed)",
    is_flag=True,
    default=False,
)
@click.option(
    "--nosuper",
    help="Don't try to create super rules that match against various files",
    is_flag=True,
    default=False,
)
@click.option(
    "-R",
    "--recursive",
    help="Recursively scan directories",
    is_flag=True,
    default=False,
)
@click.option(
    "--oe",
    "--only-executable",
    help="Only scan executable extensions EXE, DLL, ASP, JSP, PHP, BIN, INFECTED",
    is_flag=True,
    default=False,
)
@click.option(
    "-fs",
    "--max-file-size",
    help="Max file size in MB to analyze (default=2)",
    type=int,
    default=2,
)
@click.option(
    "--noextras",
    help="Don't use extras like Imphash or PE header specifics",
    is_flag=True,
    default=False,
)
@click.option("--debug", help="Debug output", is_flag=True, default=False)
@click.option("--trace", help="Trace output", is_flag=True, default=False)
@click.option(
    "--opcodes",
    help="Do use the OpCode feature (use this if not enough high scoring strings can be found)",
    is_flag=True,
    default=False,
)
@click.option(
    "-n",
    "--opcode-num",
    help="Number of opcodes to add if not enough high scoring string could be found (default=3)",
    type=int,
    default=3,
)
def generate(malware_path, **kwargs):
    pr = cProfile.Profile()
    pr.enable()
    """Generate YARA rules from malware samples"""
    args = type("Args", (), kwargs)()

    # Validate input
    if malware_path and os.path.isfile(malware_path):
        click.echo("[E] Input is a file, please use a directory instead (-m path)")
        sys.exit(0)
    sourcepath = malware_path
    args.identifier = getIdentifier(args.identifier, sourcepath)
    print("[+] Using identifier '%s'" % args.identifier)

    # Reference
    args.ref = getReference(args.ref)
    print("[+] Using reference '%s'" % args.ref)

    # Prefix
    args.prefix = getPrefix(args.prefix, args.identifier)
    print("[+] Using prefix '%s'" % args.prefix)

    pestudio_strings = initialize_pestudio_strings()
    print("[+] Reading goodware strings from database 'good-strings.db' ...")
    print(
        "    (This could take some time and uses several Gigabytes of RAM depending on your db size)"
    )

    good_strings_db, good_opcodes_db, good_imphashes_db, good_exports_db = (
        load_databases()
    )
    # exit()
    process_folder(
        args,
        malware_path,
        good_strings_db,
        good_opcodes_db,
        good_imphashes_db,
        good_exports_db,
        pestudio_strings,
    )
    pr.disable()

    stats = pstats.Stats(pr)
    stats.sort_stats("cumulative").print_stats(
        10
    )  # Sort by cumulative time and print top 10


@cli.command()
@click.argument("malware_path", type=click.Path(exists=True))
@click.option(
    "-y",
    "--min-size",
    help="Minimum string length to consider (default=8)",
    type=int,
    default=8,
)
@click.option(
    "-z",
    "--min-score",
    help="Minimum score to consider (default=5)",
    type=int,
    default=5,
)
@click.option(
    "-o", "--output-rule-file", help="Output rule file", default="yarobot_rules.yar"
)
@click.option("-a", "--author", help="Author Name", default="yarobot Rule Generator")
@click.option("--opcodes", help="Use the OpCode feature", is_flag=True, default=False)
@click.option("--debug", help="Debug output", is_flag=True, default=False)
def dropzone(malware_path, **kwargs):
    """Dropzone mode - monitor directory for new samples and generate rules automatically"""
    args = type("Args", (), kwargs)()

    click.echo(f"[+] Starting dropzone mode, monitoring {malware_path}")
    click.echo("[!] WARNING: Processed files will be deleted!")

    while True:
        if len(os.listdir(malware_path)) > 0:
            # Deactivate super rule generation if there's only a single file in the folder
            if len(os.listdir(malware_path)) < 2:
                args.nosuper = True
            else:
                args.nosuper = False
            # Read a new identifier
            identifier = getIdentifier(args.b, malware_path)
            # Read a new reference
            reference = getReference(args.ref)
            # Generate a new description prefix
            prefix = getPrefix(args.p, identifier)
            # Process the samples
            processSampleDir(malware_path)
            # Delete all samples from the dropzone folder
            emptyFolder(malware_path)
        time.sleep(1)


# MAIN ################################################################
if __name__ == "__main__":
    logging.basicConfig(level=os.environ.get("YAROBOT_LOG_LEVEL", "INFO"))
    logging.getLogger().setLevel(logging.DEBUG)
    generate()
    # Identifier
